/* $Id: gsl_util.hpp 158 2010-09-13 22:02:22Z pjohnson $
  ---------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2009. "Inference of Microbial
    Recombination Rates from Metagenomic Data."  PLoS Genetics (in
    press).
 
  Copyright 2009 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Generic wrappers for GSL functions that take a callbacks
  ---------------------------------------------------------------------------
*/
#ifndef GSL_UTIL_HPP
#define GSL_UTIL_HPP

#include <iostream>
#include <stdexcept>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_siman.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_multifit.h>
//#include <gsl/gsl_bspline.h>

namespace gsl {

//---------------------------------------------------------------------------

class CVector {
public:
    CVector(unsigned int n) { m_Vector = gsl_vector_alloc(n); }
    CVector(gsl_vector* v) {
        m_Vector = gsl_vector_alloc(v->size);
        gsl_vector_memcpy(m_Vector, v);
    }
    ~CVector(void) { gsl_vector_free(m_Vector); m_Vector = NULL; }
    unsigned int size(void) const { return m_Vector->size; }
    double get(unsigned int i) const { return gsl_vector_get(m_Vector, i); }
    friend std::ostream& operator<< (std::ostream &o, const CVector &v) {
        for (unsigned int i = 0;  i < v.size();  ++i) {
            if (i != 0) {
                o << "\t";
            }
            o << v.get(i);
        }
        return o;
    }

    CVector& operator= (const CVector &v) {
        if (size() != v.size()) {
            gsl_vector_free(m_Vector);
            m_Vector = gsl_vector_alloc(v.size());
        }
        gsl_vector_memcpy(m_Vector, v);
        return *this;
    }
    operator const gsl_vector*(void) const { return m_Vector; }
    operator gsl_vector*(void) { return m_Vector; }
private:
    CVector(CVector&);//prohibit copy construction
    gsl_vector *m_Vector;
};

//---------------------------------------------------------------------------

class CMatrix {
public:
    CMatrix(unsigned nrow, unsigned int ncol) {
        m_Matrix = gsl_matrix_alloc(nrow, ncol);
    }
    ~CMatrix(void) { gsl_matrix_free(m_Matrix); m_Matrix = NULL; }
    unsigned int nrow(void) const { return m_Matrix->size1; }
    unsigned int ncol(void) const { return m_Matrix->size2; }
    double get(unsigned int i, unsigned int j) const {
        return gsl_matrix_get(m_Matrix, i, j); }
    friend std::ostream& operator<< (std::ostream &o, const CMatrix &m) {
        for (unsigned int i = 0;  i < m.nrow();  ++i) {
            for (unsigned int j = 0;  j < m.ncol();  ++j) {
                o << (j != 0 ? "\t":"") << m.get(i,j);
            }
            if (i != m.nrow()-1) {
                o << std::endl;
            }
        }
        return o;
    }
    operator const gsl_matrix*(void) const { return m_Matrix; }
    operator gsl_matrix*(void) { return m_Matrix; }
private:
    CMatrix(CMatrix&);//prohibit copy construction
    gsl_matrix * m_Matrix;
};

//---------------------------------------------------------------------------

class Crng {
public:
    Crng(unsigned long int seed) {
        m_RNG = gsl_rng_alloc(gsl_rng_mt19937);
        gsl_rng_set(m_RNG, seed);
    }
    ~Crng(void) {
        gsl_rng_free(m_RNG);
    }

    //sampling functions
    double GetUniform(void) { return gsl_rng_uniform(m_RNG); }
    unsigned long int GetUniform(unsigned long int mx) {
        return gsl_rng_uniform_int(m_RNG, mx); }

    //for use as function adapter
    unsigned long int operator() (unsigned long int mx) {
        return GetUniform(mx); }

    operator const gsl_rng* (void) const {
        return m_RNG;
    }

private:
    gsl_rng *m_RNG;
};

//---------------------------------------------------------------------------

class CNIntegrate {
public:
    typedef CNIntegrate TThisType;
    virtual ~CNIntegrate(void) {} //virtual to silence warning

    virtual double F(double x) const = 0;
    double Integrate(double from, double to) const {
        double result, error;
        gsl_function f;
        f.function = &s_F;
        f.params = (void*) this;

        gsl_integration_workspace *w = gsl_integration_workspace_alloc(10000);
        gsl_integration_qag(&f, from, to,
                            0, 1e-7, 10000, GSL_INTEG_GAUSS15,
                            w, &result, &error);
        gsl_integration_workspace_free(w);
        return result;
    }

protected:
    //static callback functions
    static double s_F(double x, void *cls) {
        return ((const TThisType*) cls)->F(x);
    }    
};

//---------------------------------------------------------------------------

class CSimAnneal {
public:
    virtual ~CSimAnneal(void) {}
    virtual double E(void) const = 0;
    virtual void TakeStep(const gsl_rng *r, double maxSize) = 0;
    virtual double DistanceFrom(const CSimAnneal *sa) const = 0;
    virtual void PrintCfg(void) const = 0;
    virtual CSimAnneal& operator= (const CSimAnneal &sa) = 0;
    virtual CSimAnneal* Clone(void) const = 0;

    void Solve(const gsl_siman_params_t &params) {
        gsl_rng * r;
        
        gsl_rng_env_setup();
        r = gsl_rng_alloc(gsl_rng_default);
        
        gsl_siman_solve(r, this, s_E, s_Step, s_Dist, s_Print,
                        s_Copy, s_Construct, s_Destruct, 0, params);
    }

protected:
    typedef CSimAnneal TThisType;
    //static callback functions
    static double s_E(void *cls) {
        return ((const TThisType*) cls)->E();
    }
    static void s_Step(const gsl_rng *r, void *cls, double maxStep) {
        ((TThisType*) cls)->TakeStep(r, maxStep);
    }
    static double s_Dist(void *cls1, void *cls2) {
        return ((const TThisType*) cls1)->DistanceFrom((const TThisType*)cls2);
    }
    static void s_Print(void *cls) {
        ((const TThisType*) cls)->PrintCfg();
    }
    static void s_Copy(void *src, void *dst) {
        *((TThisType*) dst) = *((const TThisType*) src);
    }
    static void* s_Construct(void *cls) {
        return ((const TThisType*) cls)->Clone();
    }
    static void s_Destruct(void *cls) {
        delete ((TThisType*) cls);
    }
    //end static functions
};

//---------------------------------------------------------------------------

class CMultiMin {
public:
    typedef CMultiMin TThisType;
    typedef gsl_vector TParam;

    virtual ~CMultiMin(void) {} //virtual to silence warning
    virtual double F(const TParam*) const = 0;
    virtual void dF(const TParam *, TParam *) const {};
    virtual void FdF(const TParam *x, double *f, TParam *g) const {
        *f = F(x); dF(x, g);
    }

protected:
    //static callback functions
    static double s_F(const TParam *x, void *cls) {
        return ((const TThisType*) cls)->F(x);
    }
    static void s_dF(const TParam *x, void *cls, TParam *g) {
        ((const TThisType*) cls)->dF(x, g);
    }
    static void s_FdF(const TParam *x, void *cls, double *f, TParam *g) {
        ((const TThisType*) cls)->FdF(x, f, g);
    }
    //end static functions
};

//---------------------------------------------------------------------------

class CSimplex : CMultiMin {
protected:
    void Optimize(double stepSize, double tolerance, unsigned int maxIter) {
        CVector stepVector(x_GetNumParams());
        gsl_vector_set_all(stepVector, stepSize);
        Optimize(stepVector, tolerance, maxIter);
    }
    void Optimize(const CVector& stepVector, double tolerance,
                  unsigned int maxIter);
    bool HasConverged(void) const { return m_Converged; }

    virtual unsigned int x_GetNumParams(void) const = 0;
    virtual void x_InitializeParams(TParam*) const = 0;
    virtual void x_StoreParams(const TParam*) = 0;
private:
    bool m_Converged;
};

/*---------------------------------------------------------------------------*/
inline void CSimplex::Optimize(const CVector &stepVector, double tolerance,
                               unsigned int maxIter) {
    gsl_multimin_function func;

    func.f = &s_F;
    func.params = (void*) this;
    func.n = x_GetNumParams();

    gsl_vector *x = gsl_vector_alloc(func.n);
    x_InitializeParams(x);

    gsl_multimin_fminimizer *minimizer =
        gsl_multimin_fminimizer_alloc(gsl_multimin_fminimizer_nmsimplex,func.n);

    gsl_error_handler_t *errHandler = gsl_set_error_handler_off();
    if (gsl_multimin_fminimizer_set(minimizer, &func, x, stepVector) != 0) {
#ifdef _DEBUG
        std::cerr << "failed to start minimizer (bad initial value?)"
                  << std::endl;
#endif
        m_Converged = false;
        gsl_multimin_fminimizer_free(minimizer);
        gsl_vector_free(x);
        return;
    }
    gsl_set_error_handler(errHandler);

    int status;
    unsigned int i = 0;
    do {
        ++i;
        gsl_multimin_fminimizer_iterate(minimizer);
        status = gsl_multimin_test_size(gsl_multimin_fminimizer_size(minimizer),
                                        tolerance);
#ifdef _DEBUG
        std::cerr << "Optimization iteration #" << i << "\n";
        std::cerr << gsl_multimin_fminimizer_minimum(minimizer) << "\t";
        gsl_vector *p = gsl_multimin_fminimizer_x(minimizer);
        for (unsigned int i = 0;  i < p->size;  ++i) {
            std::cerr << gsl_vector_get(p, i) << " ";
        }
        std::cerr << std::endl;
#endif
    } while (status == GSL_CONTINUE  &&  i < maxIter);
    
    m_Converged = (status == GSL_SUCCESS);

    //assign back regardless of whether we've achieved GSL_SUCCESS
    x_StoreParams(gsl_multimin_fminimizer_x(minimizer));

    gsl_multimin_fminimizer_free(minimizer);
    gsl_vector_free(x);
}

//---------------------------------------------------------------------------

class CFunction {
public:
    typedef CFunction TThisType;

    virtual ~CFunction(void) {} //virtual to silence warning
    virtual double F(const double) const = 0;

protected:
    //static callback functions
    static double s_F(const double x, void *cls) {
        return ((const TThisType*) cls)->F(x);
    }
    //end static functions

    const gsl_function x_GetGSLFunc(void) const {
        gsl_function f;
        f.function = &s_F;
        f.params = (void*) this;
        return f;
    }
};

class CDFunction {
public:
    typedef CDFunction TThisType;

    virtual ~CDFunction(void) {} //virtual to silence warning
    virtual double F(const double) const = 0;
    virtual double dF(const double) const = 0;
    virtual void FdF(const double x, double *f, double *df) const {
        *f = F(x); *df=dF(x);
    }

protected:
    //static callback functions
    static double s_F(const double x, void *cls) {
        return ((const TThisType*) cls)->F(x);
    }
    static double s_dF(const double x, void *cls) {
        return ((const TThisType*) cls)->dF(x);
    }
    static void s_FdF(const double x, void *cls, double *f, double *df) {
        ((const TThisType*) cls)->FdF(x, f, df);
    }
    //end static functions

    const gsl_function_fdf x_GetGSLDFunc(void) const {
        gsl_function_fdf f;
        f.f = &s_F;
        f.df = &s_dF;
        f.fdf = &s_FdF;
        f.params = (void*) this;
        return f;
    }
};

//---------------------------------------------------------------------------

class CFindRootBracket : public CFunction {
protected:
    double FindRoot(double lowerBound, double upperBound,
                    double tolerance, unsigned int maxIter)
    {
#ifdef _DEBUG
        std::cerr << "lb: " << lowerBound << "\tub: " << upperBound << std::endl;
#endif
        gsl_function func(x_GetGSLFunc());
        gsl_root_fsolver *solver =
            gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
        gsl_root_fsolver_set(solver, &func, lowerBound, upperBound);

        int status;
        unsigned int i = 0;
        do {
            gsl_root_fsolver_iterate(solver);
            status = gsl_root_test_interval(gsl_root_fsolver_x_lower(solver),
                                            gsl_root_fsolver_x_upper(solver),
                                            0, tolerance);
        } while(status == GSL_CONTINUE  &&  ++i < maxIter);

        m_Converged = (status == GSL_SUCCESS);

        double root = gsl_root_fsolver_root(solver);
        gsl_root_fsolver_free(solver);

        return root;
    }

    bool HasConverged(void) const { return m_Converged; }

private:
    bool m_Converged;
};

//---------------------------------------------------------------------------

class CFindRootDerive : public CDFunction {
public:
    enum ESolver {eNewtonRaphson, eSteffenson};
protected:
    double FindRoot(double guess,
                    double tolerance, unsigned int maxIter,
                    ESolver solveType = eNewtonRaphson)
    {
        gsl_function_fdf func(x_GetGSLDFunc());
        gsl_root_fdfsolver *solver;
        switch (solveType) {
        case eNewtonRaphson:
            solver = gsl_root_fdfsolver_alloc(gsl_root_fdfsolver_newton);
            break;
        case eSteffenson:
            solver = gsl_root_fdfsolver_alloc(gsl_root_fdfsolver_steffenson);
            break;
        }            
        gsl_root_fdfsolver_set(solver, &func, guess);

        double prevRoot, root;
        int status;
        gsl_root_fdfsolver_iterate(solver);
        root = gsl_root_fdfsolver_root(solver);
        unsigned int i = 1;
        do {
            gsl_root_fdfsolver_iterate(solver);
            prevRoot = root;
            root = gsl_root_fdfsolver_root(solver);
            status = gsl_root_test_delta(root, prevRoot, 0, tolerance);

        } while(status == GSL_CONTINUE  &&  ++i < maxIter);

        m_Converged = (status == GSL_SUCCESS);

        gsl_root_fdfsolver_free(solver);

        return root;
    }
    bool HasConverged(void) const { return m_Converged; }

private:
    bool m_Converged;
};

//---------------------------------------------------------------------------

class CMultiFit {
public:
    static double FitData(const CMatrix &x, const CVector &y, CVector &coefs) {
        double chisq;
        unsigned int numObs = x.nrow();
        unsigned int numCoefs = x.ncol();

        gsl_matrix *cov      = gsl_matrix_alloc(numCoefs, numCoefs);

        gsl_multifit_linear_workspace *work =
            gsl_multifit_linear_alloc(numObs, numCoefs);
        gsl_multifit_linear(x, y, coefs, cov, &chisq, work);
        gsl_multifit_linear_free(work);

        gsl_matrix_free(cov);
        return chisq;
    }
};

//---------------------------------------------------------------------------
/*
class CBSpline {
public:
    // order 4 == cubic spline
    CBSpline (const CVector& x, const CVector& y,
              size_t nbreak, size_t order = 4) :
        m_Work(gsl_bspline_alloc(order, nbreak)),
        m_Coefs(gsl_bspline_ncoeffs(m_Work))
    {
        if (x.size() != y.size()) {
            throw std::runtime_error("CBSpline passed x and y vectors of "
                                     "different length.");
        }
        
        double minX, maxX;
        minX = maxX = x.get(0);
        for (unsigned int i = 1;  i < x.size();  ++i) {
            if (x.get(i) < minX) { minX = x.get(i); }
            if (x.get(i) > maxX) { maxX = x.get(i); }
        }
        gsl_bspline_knots_uniform(minX, maxX, m_Work);

        CVector basisVector(m_Coefs.size());
        CMatrix fitMatrix(x.size(), m_Coefs.size());
    
        for (unsigned int i = 0;  i < x.size();  ++i) {
            gsl_bspline_eval(x.get(i), basisVector, m_Work);
            gsl_matrix_set_row(fitMatrix, i, basisVector);
        }
        CMultiFit::FitData(fitMatrix, y, m_Coefs);
    }
    ~CBSpline(void) {
        gsl_bspline_free(m_Work);
    }

    double EvalAt(double x) const {
        CVector basisVector(gsl_bspline_ncoeffs(m_Work));
        gsl_bspline_eval(x, basisVector, m_Work);
        double res = 0;
        for (unsigned int i = 0;  i < m_Coefs.size();  ++i) {
            res += m_Coefs.get(i) * basisVector.get(i);
        }
        return res;
    }

private:
    gsl_bspline_workspace *m_Work;
    CVector m_Coefs;
};
*/
//---------------------------------------------------------------------------

} //end gsl namespace

#endif /*ifndef GSL_UTIL_HPP*/
