/* $Id: piim_data.cpp 358 2020-03-16 23:03:27Z pjohnson $
  ---------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2009. "Inference of Microbial
    Recombination Rates from Metagenomic Data."  PLoS Genetics (in
    press).
 
  Copyright 2009 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Classes for handling metagenomic data with PIIM
  ---------------------------------------------------------------------------
*/

#include "piim_data.hpp"
#include "util.hpp"
#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>
#include <memory>

using namespace std;

/*---------------------------------------------------------------------------*/

struct XML { //don't pollute global namespace
    static int x_StreamRead(void * i, char* buff, int len) {
        ((istream*)i)->read(buff, len);
        return ((istream*)i)->gcount();
    }
    static int x_StreamWrite(void * i, const char* buff, int len) {
        ((ostream*)i)->write(buff, len);
        return ((ostream*)i)->good() ? len:-1;
    }
    static void x_ErrorHandler(void *, const char *msg, xmlParserSeverities,
                               xmlTextReaderLocatorPtr) {
        throw runtime_error("XML reader says "+string(msg));
    }

    //hacky, but works (at least on my linux box)
    class string : public std::string {
    public:
        string(xmlChar* c) : std::string(c ? (const char*) c : ""), m_Str(c) {}
        string(const char* c) : std::string(c), m_Str(xmlCharStrdup(c)) {}
        string(const std::string &c) : std::string(c),
                                       m_Str(xmlCharStrdup(c.c_str())) {}
        ~string(void) {xmlFree(m_Str);}
        operator xmlChar const *(void) {return m_Str;}
    private:
        xmlChar *m_Str;
    };
    enum xNodeTypes {eElementStart = 1, eElementEnd = 15, eText = 3};
};

/*---------------------------------------------------------------------------*/

xmlTextWriterPtr operator<< (xmlTextWriterPtr xml, const CRead &read) {
    xmlTextWriterStartElement(xml, (const xmlChar*) "Read");
    xmlTextWriterWriteAttribute(xml, (const xmlChar*) "start",
                                (const xmlChar*) IntToString(read.GetStart()).c_str());
    xmlTextWriterWriteAttribute(xml, (const xmlChar*) "acc",
                                XML::string(read.GetAcc()));
    {
        string baseTxt, qualTxt;
        baseTxt.reserve(read.GetLength());
        qualTxt.reserve(read.GetLength());
        for (unsigned int i = read.GetStart();  i <= read.GetStop();  ++i) {
            baseTxt += Enum2Ch(read.GetBase(i));
            qualTxt += IntToString(read.GetQual(i)) + " ";
        }

        xmlTextWriterWriteElement(xml, XML::string("Bases"), XML::string(baseTxt));
        xmlTextWriterWriteElement(xml, XML::string("Quals"), XML::string(qualTxt));
    }
    xmlTextWriterEndElement(xml);
    return xml;
}

/*---------------------------------------------------------------------------*/

xmlTextWriterPtr operator<< (xmlTextWriterPtr xml, const CChr &chr) {
    xmlTextWriterStartElement(xml, XML::string("Chr"));
    for (CChr::TReads::const_iterator readI = chr.GetReads().begin();
         readI != chr.GetReads().end();  ++readI) {
        xml << *readI;
    }
    xmlTextWriterEndElement(xml);
    return xml;
}

/*---------------------------------------------------------------------------*/
// PRE : open output stream, assembly
// POST: XML-formatted assembly saved to stream
void CAssembly::Save(ostream &o) const {
    xmlTextWriterPtr xml =
        xmlNewTextWriter(xmlOutputBufferCreateIO
                         (&XML::x_StreamWrite, NULL, &o, NULL));
    xmlTextWriterSetIndent(xml, 1);
    xmlTextWriterStartDocument(xml, NULL, NULL, NULL);
    xmlTextWriterWriteDTD(xml, XML::string("Contig"), NULL,
                          XML::string("piim_r.dtd"), NULL);


    xmlTextWriterStartElement(xml, XML::string("Contig"));
    for (unsigned int i = 0;  i < GetNumChrs();  ++i) {
        xml << (*this)[i];
    }
    xmlTextWriterEndElement(xml);

    xmlFreeTextWriter(xml);    
}

/*---------------------------------------------------------------------------*/
// PRE : open input stream with assembly in XML format
// POST: assembly loaded into memory
void CAssembly::Load(istream &in) {
    unique_ptr<CChr> chr;
    unique_ptr<CRead> read;
    xmlChar *textValue = NULL;

    xmlTextReaderPtr xml =
        xmlReaderForIO(&XML::x_StreamRead, NULL, &in, NULL, NULL,
                       XML_PARSE_DTDVALID);
    xmlTextReaderSetErrorHandler(xml, &XML::x_ErrorHandler, NULL);

    while (xmlTextReaderRead(xml) == 1) {
        if (xmlTextReaderNodeType(xml) == XML::eElementStart) {
            XML::string name (xmlTextReaderName(xml));
            if (name == "Chr") {
                chr.reset(new CChr);

            } else if (name == "Read") {
                XML::string s(xmlTextReaderGetAttribute
                              (xml, XML::string("start")));
                if (!s) {
                    throw runtime_error("Missing 'start' attribute in <READ>");
                }
                read.reset(new CRead(atoi(s.c_str())));
                XML::string acc(xmlTextReaderGetAttribute
                                (xml, XML::string("acc")));
                read->SetAcc(acc);
            }
        } else if (xmlTextReaderNodeType(xml) == XML::eText) {
            xmlFree(textValue);
            textValue = xmlTextReaderValue(xml);
        } else if (xmlTextReaderNodeType(xml) == XML::eElementEnd) {
            XML::string name (xmlTextReaderName(xml));
            if (name == "Chr") {
                //for speed, call directly WITHOUT merging; sort at end
                m_Chrs.push_back(chr.release());//AddChr(chr.release());

            } else if (name == "Read") {
                chr->AddRead(*read);

            } else if (name == "Bases") {
                CRead::TBases &bases = read->SetBases();
                bases.resize(xmlStrlen(textValue));
                unsigned int i = 0;
                for (xmlChar *p = textValue;  *p != '\0';  ++p) {
                    bases[i] = Ch2Enum(*p);
                    ++i;
                }
                
            } else if (name == "Quals") {
                CRead::TQuals &quals = read->SetQuals();
                quals.resize(read->SetBases().size());
                unsigned int i = 0;

                //numeric quals or FASTQ-style quals?
                if ((int) quals.size() != xmlStrlen(textValue)  ||
                    //assume numeric
                    xmlStrchr(textValue, ' ')) {
                    istringstream iss;
                    iss.str((const char*)textValue);

                    unsigned short int q;
                    while (!iss.eof()  &&  i < quals.size()) {
                        iss >> q;
                        if (iss.fail()) {
                            throw runtime_error("Invalid/missing q-score value "
                                                "in read '" + read->GetAcc() +
                                                "'!");
                        }
                        quals[i] = q;
                        ++i;
                    }
                } else if ((int) quals.size() == xmlStrlen(textValue)) {
                    //assume fastq-style (ascii-33)
                    for (xmlChar *p = textValue;  *p != '\0';  ++p) {
                        quals[i] = (unsigned char) *p - 33;
                        ++i;
                    }
                }

                if (i != quals.size()) {
                    throw runtime_error("Different number of bases and q-scores for read '" + read->GetAcc() + "'!");
                }
            }
        }

    }
    bool valid = xmlTextReaderIsValid(xml);
    xmlFree(textValue);
    xmlFreeTextReader(xml);
    if (!valid) {
        throw runtime_error("XML validation failed.");
    }
    sort(m_Chrs.begin(), m_Chrs.end(), CChr::CPtrCmp());
}
